/*****************************************************
Compiler            : CodeVisionAVR 2.04.6 Evaluation
Chip type           : ATmega8
Clock frequency     : 8 MHz (int. RC-OSC)
*****************************************************/

//***************************************************
// Elektronische Spardose
// auf Basis des Franzis-Pingpong
// http://www.elo-web.de/ping-pong-start
//
// Achtung: bei Project>Configure>c-COmpiler muss bei
// der Option "Clear Global Variables at Startup"
// der Haken entfernt werden! Grund: Da der Watchdog-Timer-Reset
// verwendet wird, wird eine Variable bentigt, die
// nicht durch einen Reset vorbelegt wird. 
//***************************************************

#include <mega8.h>
#include <stdio.h>
#include <delay.h>
#include <sleep.h>


#define P_CLK    PORTB.3
#define P_DATA   PORTB.4
#define P_STROBE PORTB.2

#define PIN_COIN     PIND.2
#define PULLUP_COIN  PORTD.2
#define DIR_COIN     DDRD.2
#define COIN_INSERTED (PIN_COIN == 0)

#define WIDTH  12  // number of fields in horizontal direction
#define HEIGHT 10  // number of fields in vertical direction
unsigned int leds[WIDTH];  // current state of each LED (organized in columns)
bit bBlink;                // blinking display

unsigned char cWakeup;  // Attention: must not be initialized!
                        // Needs to keep value after watchdog reset!


eeprom unsigned int eeSum = 0;   // coin sum is stored in EEPROM

flash unsigned char coinValue[] = { 1, 2, 5, 10, 20, 50, 100, 200 };  // 1 Cent ... 2 
enum { CENT1 = 0, CENT2, CENT5, CENT10, CENT20, CENT50, EURO1, EURO2, SUM_RIGHT, SUM_LEFT };


flash unsigned int logoEuro[] = { 0x0177, 0x0151, 0x0157, 0x0151, 0x0357 }; // EUR

flash unsigned int logoCent[] = { 0x02BB, 0x0289, 0x02B9, 0x0289, 0x03BB }; // CEN

flash unsigned char decimals[] =
{ 
0x7, 0x5, 0x5, 0x5, 0x7,   // 0
0x4, 0x4, 0x5, 0x6, 0x4,   // 1
0x7, 0x1, 0x7, 0x4, 0x7,   // 2
0x7, 0x4, 0x7, 0x4, 0x7,   // 3
0x4, 0x4, 0x7, 0x5, 0x5,   // 4
0x7, 0x4, 0x7, 0x1, 0x7,   // 5
0x7, 0x5, 0x7, 0x1, 0x7,   // 6
0x4, 0x4, 0x4, 0x4, 0x7,   // 7
0x7, 0x5, 0x7, 0x5, 0x7,   // 8
0x7, 0x4, 0x7, 0x5, 0x7,   // 9
0x6, 0x1, 0x7, 0x1, 0x6    // EURO symbol 
};


// clear the whole screen, all LEDs off
void clearScreen(void)
{
   unsigned char i;
   
   for(i = 0; i < (WIDTH); i++)
   {
      leds[i] = 0;
   }
}


// set controller in Powerdown mode, wakeup is done via Watchdog reset after 1 second
void setStandbyMode(void)
{
   TCCR2 &= 0xF8;   //stop Timer2
   PORTC = 0;
   PORTD = 0;
   PORTB = 0;
   ADCSRA &= 0x7F;  // stopADC
   DDRD.2 = 0;
   PORTD.2 = 1;
   
   // Watchdog Timer initialization
   // Watchdog Timer Prescaler: OSC/1024k  ==> WDT reset after 1 second
   #pragma optsize-
   WDTCR=0x1E;
   WDTCR=0x0E;
   #ifdef _OPTIMIZE_SIZE_
   #pragma optsize+
   #endif

   sleep_enable();
   powerdown();
}


// Timer 2 overflow interrupt service routine
// multiplexing of LEDs
interrupt [TIM2_OVF] void timer2_ovf_isr(void)
{
   static unsigned char count = 0;
   unsigned int iValue;
   static unsigned int iBlinkMask = 0xFFFF;
   static unsigned char cBlinkcnt = 0;
   
   count++;
   if(count >= 12)
   {
      count = 0;
      P_DATA = 0;
      cBlinkcnt++;
      if(cBlinkcnt == 81)
      {
         cBlinkcnt = 0;
         iBlinkMask = ~iBlinkMask;
      }
   }
   else
   {
      P_DATA = 1;
   }
   P_CLK = 1;
   P_CLK = 0;
   PORTC &= (~0x0F);
   PORTD &= (~0xF0);
   PORTB &= (~0x03);
   P_STROBE = 1;
   P_STROBE = 0;
   iValue = leds[count];
   if(bBlink)
      iValue &= iBlinkMask; 
   PORTC |= (iValue & 0x0F);
   PORTD |= (iValue & 0xF0);
   PORTB |= ((iValue >> 8) & 0x03);
   
}


// initialization of Timer2
void initTimer2(void)
{
   // Timer/Counter 2 initialization
   // Clock source: System Clock
   // Clock value: 1000,000 kHz
   // Mode: Normal top=FFh
   // OC2 output: Disconnected
   ASSR=0x00;
   TCCR2=0x02;
   TCNT2=0x00;
   OCR2=0x00;

   // Timer(s)/Counter(s) Interrupt(s) initialization
   TIMSK=0x40;
}


// initialization of ports
void init(void)
{
   PORTB = 0;
   DDRB = 0x1F;
   PORTC = 0;
   DDRC = 0x0F;
   PORTD = 0;
   DDRD = 0xF4;

   // Analog Comparator initialization
   // Analog Comparator: Off
   // Analog Comparator Input Capture by Timer/Counter 1: Off
   ACSR=0x80;
   SFIOR=0x00;
}



#define ADC_VREF_TYPE 0x00

// initialization of ADC
void initADC(void)
{
   // ADC initialization
   // ADC Clock frequency: 125,000 kHz
   // ADC Voltage Reference: AREF pin
   ADMUX=ADC_VREF_TYPE & 0xff;
   ADCSRA=0x86;
}


// Read the AD conversion result
// adc_input: ADC channel (0..7)
// return: ADC value (0..1023)
unsigned int read_adc(unsigned char adc_input)
{
   ADMUX=adc_input | (ADC_VREF_TYPE & 0xff);
   // Delay needed for the stabilization of the ADC input voltage
   delay_us(10);
   // Start the AD conversion
   ADCSRA|=0x40;
   // Wait for the AD conversion to complete
   while ((ADCSRA & 0x10)==0);
   ADCSRA|=0x10;
   
   return ADCW;
}


// read value of Poti2 (used for clearing coin sum)
// return: 0..9
unsigned char poti2(void)
{
   unsigned char x;                        
   
   x = (read_adc(7) / 103);
   
   return(x);
}



// select coin to be inserted
// return: selected coin index or sum index (CENT1...EURO2, SUM_RIGHT,SUM_LEFT)
unsigned char selectCoin(void)
{
   unsigned char x;                        
   
   x = (read_adc(6) / 103);
   if(x == 0)
      x = SUM_LEFT;
   else if(x == 9)
      x = SUM_RIGHT;
   else
      x--;
   
   return(x);
}


// display current sum of all coins
// sum: sum to be displayed
void displaySum(unsigned int sum)
{
   unsigned char i, digit, value;
   unsigned int led;
   
      
   leds[0] = 0;   
   
   sum %= 10000;
   
   value = sum % 100;      // display cents
   for(i = 0; i < 5; i++)
   {
      digit = value / 10;
      led = decimals[digit*5 + i];
      digit = value % 10;
      led |= (decimals[digit*5 + i] << 4);
      digit = 10;    // EURO symbol 
      led |= ( ((unsigned int)decimals[digit*5 + i] & 0x03) << 8 );
      leds[i+1] = led;
   }

   leds[6] = 0;
   
   value = sum / 100;      // display Euros
   value %= 100;
   for(i = 0; i < 5; i++)
   {
      digit = value / 10;
      if(digit != 0)
         led = decimals[digit*5 + i];
      else
         led = 0;               // suppress leading zeros
      if(i == 0)
         led |= 0x100;  // decimal point
      digit = value % 10;
      led |= (decimals[digit*5 + i] << 4);
      leds[i+7] = led;
   }
}


// display coin value
// x: coin (CENT1..EURO2)
void displayCoinValue(unsigned char x)
{
   unsigned char i, digit, value;
   unsigned int led;
   flash unsigned int *pLogo;
   
   if(x < EURO1)
   {
      pLogo = logoCent;
      value = coinValue[x];
   }
   else
   {
      pLogo = logoEuro;
      value = coinValue[x] / 100;
   }
      
   leds[0] = 0;   
   for(i = 0; i < 5; i++)
   {
      leds[i+1] = pLogo[i];
   }

   leds[6] = 0;
   
   for(i = 0; i < 5; i++)
   {
      digit = value / 10;
      if(digit != 0)
      {                 
         led = decimals[digit*5 + i];
      }
      else
      {
         led = 0;
      }
      digit = value % 10;
      led |= (decimals[digit*5 + i] << 4);
      leds[i+7] = led;
   }
   
}


// main routine
void main(void)
{
   unsigned char x, lastX, timeout, p2, cDelete;
   unsigned int sum;
   bit bReady, bCoinInserted;
   bit bWakeup;

   init();
   initTimer2();
   initADC();
               
   bWakeup = 0;
   x = selectCoin();

   if( (cWakeup != SUM_LEFT) && (cWakeup != SUM_RIGHT) && (cWakeup != 0xFF) )
      cWakeup = 0xFF;  // if value is undefined at power-on reset, right or left position of Poti1 is possible for wakeup
   
   if( (cWakeup == SUM_LEFT) || (cWakeup == 0xFF) )
   {
      if(x == SUM_LEFT)
         bWakeup = 1;
   }                 
   if( (cWakeup == SUM_RIGHT) || (cWakeup == 0xFF) )
   {
      if(x == SUM_RIGHT)
         bWakeup = 1;
   }                 
   
   if(!bWakeup)
   {  // if Poti1 is not in correct position => sleep again until next check in 1 second
      setStandbyMode();
   }


   clearScreen();
   bBlink = 0;
   cDelete = 0;
   
   #asm("sei")      // Global enable interrupts

   
   while(1)
   {                  
      timeout = 250;
      bReady = 0;
      bCoinInserted = 0;
      do
      {
         PULLUP_COIN = 0;
         DIR_COIN = 1;
         delay_ms(20);

         x = selectCoin();
         if(x < SUM_RIGHT)
         {
            displayCoinValue(x);
            cDelete = 0;
            bBlink = 0;
         }
         else
         {
            sum = eeSum;
            displaySum(sum);
            p2 = poti2();
            if(p2 == 0)
               cDelete |= 0x01;
            else if(p2 == 9)
               cDelete |= 0x02;
            if(cDelete)
               bBlink = 1;
            if((cDelete & 0x03) == 0x03)
            {
               sum = 0;
               eeSum = sum;
               cDelete = 0;
               bReady = 1;
            }
         }

         PULLUP_COIN = 1;
         DIR_COIN = 0;
         delay_ms(20);
         timeout--;
         
         if(COIN_INSERTED)
         {             
            bCoinInserted = 1;
            bReady = 1;
         }
         if(lastX != x)
         {         
            bReady = 1;
         }
         lastX = x;
      }
      while( (!bReady) && (timeout != 0) );

      if(timeout == 0)
      {            
         if( (x == 3) || (x == 4) )
            cWakeup = 0xFF;
         else if( (x == SUM_LEFT) || (x < 3) )
            cWakeup = SUM_RIGHT;
         else
            cWakeup = SUM_LEFT;
             
         setStandbyMode();
      }
      else
      {
         if(bCoinInserted && (x <= EURO2))
         {            
            sum = eeSum;
            sum += coinValue[x];
            eeSum = sum;
            displaySum(sum);
            delay_ms(2000);
         }
      }

   }   
}
